/*
 * fsm.c - Control Finite State Machine for diald main loop.
 *
 * Copyright (c) 1994 Eric Schenk.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL ERIC SCHENK BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF ERIC
 * SCHENK HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ERIC SCHENK SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND ERIC SCHENK HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 */
#include "diald.h"
#include "fsm.h"

#define STATE(name,timeout,tstate) \
        { timeout, act_ ## name, trans_ ## name, STATE_ ## tstate, #name }


int blocked = 0;		/* user has blocked the link */
int state = STATE_DOWN;		/* DFA state */
int current_retry_count = 0;	/* current retry count */
int no_redial_delay = 0;

void GOTO(int);

/*
 * Actions and transition rules for the FSM.
 */

void act_DOWN(void)
{
    current_retry_count = retry_count;
}
void trans_DOWN(void)
{
    request_down = 0;
    if (request_up && !blocked) {request_up = 0; GOTO(STATE_CONNECT);}
    else if (!queue_empty() && !blocked) GOTO(STATE_CONNECT);
}

/* Grumble. Attempt to generate a nicely formatted ascii date without
 * a built in newline.
 */
char *cdate(void)
{
    static char dt[128];
    int len;

    time((time_t *)&call_start_time);
    len = strftime(dt,128,"%c %Z",localtime((time_t *)&call_start_time));
    if (len == 128) dt[len] = 0;
    return dt;
}

void act_CONNECT(void)
{
    if (acctfp) {
	fprintf(acctfp,"%s: Calling site %s.\n",
	    cdate(), remote_ip);
	fflush(acctfp);
    }
    txtotal = rxtotal = 0;
    open_modem();
}
void trans_CONNECT(void)
{
    if (dial_pid == 0)
	if (dial_status == 0) {
	    finish_dial(); /* go into a CLOCAL connection */
	    GOTO(STATE_START_LINK);
	} else {
	    syslog(LOG_INFO,"Connect script failed.");
	    GOTO(STATE_CLOSE);
	}
}

void act_STOP_DIAL(void)
{
    if (dial_pid) {
        if (debug&DEBUG_VERBOSE)
            syslog(LOG_INFO,"Sending SIGINT to (dis)connect process %d",
		dial_pid);
        kill(dial_pid,SIGINT);
    }
}
void trans_STOP_DIAL(void)
{
    if (dial_pid == 0) GOTO(STATE_CLOSE);
}


void act_KILL_DIAL(void)
{
    if (dial_pid) {
        if (debug&DEBUG_VERBOSE)
            syslog(LOG_INFO,"Sending SIGKILL to (dis)connect process %d",
		dial_pid);
        kill(dial_pid,SIGKILL);
    }
}
void trans_KILL_DIAL(void)
{
    if (dial_pid == 0) GOTO(STATE_CLOSE);
}

void act_START_LINK(void)
{
    ppp_iface = -1;
    if (acctfp) {
        fprintf(acctfp,"%s: Connected to site %s.\n",
	    cdate(), remote_ip);
	fflush(acctfp);
    }
    if (mode == MODE_PPP) fork_pppd();
    else if (dynamic_addrs) dynamic_slip();
}
void trans_START_LINK(void)
{
    if (mode == MODE_SLIP) GOTO(STATE_UP);
    else if (pppd_pid == 0) GOTO(STATE_DISCONNECT);
    else if (find_ppp_iface()) GOTO(STATE_UP);
}

void act_STOP_PPPD(void)
{
    if (pppd_pid) kill(pppd_pid,SIGINT);
}
void trans_STOP_PPPD(void)
{
    if (pppd_pid == 0) GOTO(STATE_DISCONNECT);
}

void act_KILL_PPPD(void)
{
    if (pppd_pid) kill(pppd_pid,SIGKILL);
}
void trans_KILL_PPPD(void)
{
    if (pppd_pid == 0) GOTO(STATE_DISCONNECT);
}

void act_UP(void)
{
    idle_filter_init();
}
void trans_UP(void)
{
    request_up = 0;
    if (blocked || request_down) {
	request_down = 0;
	goto take_link_down;
    }
    if (state_timeout == 0) {
	syslog(LOG_INFO,"%s %d %s",
	    "Failed to received first packet within",
	    first_packet_timeout,
	    "seconds. Closing Link down.");
	goto take_link_down;
    }
    if (queue_empty() && state_timeout == -1) {
	syslog(LOG_INFO,"Closing down idle link.");
take_link_down:
	current_retry_count = 0;
	flush_timeout_queue();
	idle_filter_proxy();
	if (mode == MODE_PPP) GOTO(STATE_STOP_PPPD);
	else GOTO(STATE_DISCONNECT);
    }
    if ((mode == MODE_PPP && pppd_pid == 0) || modem_hup) {
	no_redial_delay = 1;
	current_retry_count = died_retry_count;
	idle_filter_proxy();
	if (modem_hup) GOTO(STATE_CLOSE);
	else GOTO(STATE_DISCONNECT);
    }
}

void act_DISCONNECT(void)
{
   dial_status = 0;
   if (disconnector) {
       fork_dialer(disconnector, modem_fd, modem_fd);
   }
}
void trans_DISCONNECT(void)
{
    if (dial_pid == 0) {
	if (dial_status != 0)
	    syslog(LOG_INFO,"Disconnect script failed");
	GOTO(STATE_CLOSE);
    }
}

void act_CLOSE(void)
{
    if (acctfp) {
	int duration = -call_start_time;
	char *tm;
	tm = cdate();
	duration += call_start_time;
        fprintf(acctfp,"%s: Disconnected. Call duration %d seconds.\n",
	    tm,duration);
        fprintf(acctfp,"      IP transmited %d bytes and received %d bytes.\n",
	    txtotal,rxtotal);
	fflush(acctfp);
    }
    close_modem();
    if (!no_redial_delay)
        syslog(LOG_INFO,"Delaying %d seconds before clear to dial.",
	    redial_timeout);
}
void trans_CLOSE(void)
{
    if (no_redial_delay == 1) {
	no_redial_delay = 0;
	GOTO(STATE_RETRY);
    }
}

void act_RETRY(void)
{
    current_retry_count--;
}
void trans_RETRY(void)
{
    if (current_retry_count >= 0) GOTO(STATE_CONNECT);
    else GOTO(STATE_DOWN);
}


void act_ERROR(void)
{
    syslog(LOG_ERR,"Subprocess sent SIGKILL still alive after %d seconds. diald quiting.",kill_timeout);
    proxy_down();
    exit(1);
}
void trans_ERROR(void) {}

/*
 * State definitions and actions for the control DFA.
 * The routine change_state handles transitions.
 */

struct {
    int *timeout;
    void (*action)(void);
    void (*trans)(void);
    int timeout_state;
    char *name;
} trans[] = {
    STATE(DOWN ,0, DOWN),
    STATE(CONNECT, &connect_timeout, STOP_DIAL),
    STATE(STOP_DIAL, &stop_dial_timeout, KILL_DIAL),
    STATE(KILL_DIAL, &kill_timeout, ERROR),
    STATE(START_LINK, &start_pppd_timeout, STOP_PPPD),
    STATE(STOP_PPPD, &stop_pppd_timeout, KILL_PPPD),
    STATE(KILL_PPPD, &kill_timeout, ERROR),
    STATE(UP, &first_packet_timeout, UP),
    STATE(DISCONNECT, &disconnect_timeout, STOP_DIAL),
    STATE(CLOSE, &redial_timeout, RETRY),
    STATE(RETRY, 0, RETRY),
    STATE(ERROR, 0, ERROR)
};

void GOTO(int new_state)
{
    state_timeout = (trans[new_state].timeout)
                     ? *trans[new_state].timeout : -1;
    if (debug&DEBUG_STATE_CONTROL)
    	syslog(LOG_DEBUG,"new state %s action %d timeout %d",
	    trans[new_state].name,trans[new_state].action,state_timeout);
    state = new_state;
    (*trans[new_state].action)();
}


/*
 * Change the state according to the current state and events.
 */

void change_state()
{
    /* Change state from the current state if conditions warrent */
    (*trans[state].trans)();
    /* Check if the current state timed out */
    if (state_timeout == 0) GOTO(trans[state].timeout_state);
}
